class BaseType:
    def __init__(self, scopes):
        for x in scopes:
            self.__setattr__(x[0], x[1])
        self._view = self.__dict__.copy()

    def __repr__(self):
        return f"{self.__class__.__name__}(" + ", ".join([f"{x[0]}={x[1]}" for x in self._view.items() if x[1] is not None]) + ")"


class User(BaseType):
    def __init__(self,
                 id,
                 is_bot,
                 first_name,
                 last_name=None,
                 username=None,
                 language_code=None,
                 can_join_groups=None,
                 can_read_all_group_messages=None,
                 supports_inline_queries=None):
        super().__init__(list(locals().items())[1:])


class Chat(BaseType):
    def __init__(self,
                 id,
                 type,
                 title=None,
                 username=None,
                 first_name=None,
                 last_name=None,
                 photo=None,
                 bio=None,
                 has_private_forwards=None,
                 description=None,
                 invite_link=None,
                 pinned_message=None,
                 permissions=None,
                 slow_mode_delay=None,
                 message_auto_delete_time=None,
                 has_protected_content=None,
                 sticker_set_name=None,
                 can_set_sticker_set=None,
                 linked_chat_id=None,
                 location=None):
        super().__init__(list(locals().items())[1:])


class Message(BaseType):
    def __init__(self,
                 message_id,
                 date,
                 chat,
                 from_=None,
                 sender_chat=None,
                 forward_from=None,
                 forward_from_chat=None,
                 forward_from_message_id=None,
                 forward_signature=None,
                 forward_sender_name=None,
                 forward_date=None,
                 is_automatic_forward=None,
                 reply_to_message=None,
                 via_bot=None,
                 edit_date=None,
                 has_protected_content=None,
                 media_group_id=None,
                 author_signature=None,
                 text=None,
                 entities=None,
                 animation=None,
                 audio=None,
                 document=None,
                 photo=None,
                 sticker=None,
                 video=None,
                 video_note=None,
                 voice=None,
                 caption=None,
                 caption_entities=None,
                 contact=None,
                 dice=None,
                 game=None,
                 poll=None,
                 venue=None,
                 location=None,
                 new_chat_members=None,
                 left_chat_member=None,
                 new_chat_title=None,
                 new_chat_photo=None,
                 delete_chat_photo=None,
                 group_chat_created=None,
                 supergroup_chat_created=None,
                 channel_chat_created=None,
                 message_auto_delete_timer_changed=None,
                 migrate_to_chat_id=None,
                 migrate_from_chat_id=None,
                 pinned_message=None,
                 invoice=None,
                 successful_payment=None,
                 connected_website=None,
                 passport_data=None,
                 proximity_alert_triggered=None,
                 voice_chat_scheduled=None,
                 voice_chat_started=None,
                 voice_chat_ended=None,
                 voice_chat_participants_invited=None,
                 reply_markup=None):
        super().__init__(list(locals().items())[1:])




import requests


def make_request(url, json=None, headers=None, method="post"):
    if method.lower() == "post":
        return requests.post(url, json=json, headers=headers)
    elif method.lower() == "get":
        return requests.get(url, json=json, headers=headers)




import websocket
import json
import datetime
from ..make_request import make_request
from silverbots.discord.types import *


def convert_date(date):
    return int(datetime.datetime.strptime(date.split(".")[0], "%Y-%m-%dT%H:%M:%S").replace(tzinfo=datetime.timezone.utc).timestamp())


def convert_helper(req, to_class):
    if to_class == User:
        if "bot" in req:
            is_bot = req["bot"]
        else:
            is_bot = False
        return User(id=req["id"], is_bot=is_bot, first_name=req["username"])
    elif to_class == Chat:
        return Chat(id=req["channel_id"], type="channel")
    elif to_class == Message:
        return Message(message_id=req["id"], date=convert_date(req["timestamp"]), chat=Chat(id=req["channel_id"], type="channel"), from_=convert_helper(req["author"], User), text=req["content"])


class DiscordBot:
    def __init__(self,
                 token):
        self.token = token
        self.headers = {"Authorization": f"Bot {token}"}
        self.web = "https://discord.com/api/v9/"
        self.id = self.get_me().id
        self.handle_message_funcs = []

    def handle_message(self):
        def wrapped(wrp):
            self.handle_message_funcs.append(wrp)
        return wrapped

    def get_me(self) -> User:
        return convert_helper(make_request(self.web + "users/@me", headers=self.headers, method="GET").json(), User)

    def send_message(self, chat_id, text) -> Message:
        return convert_helper(make_request(self.web + f"channels/{chat_id}/messages", json={"content": text}, headers=self.headers, method="POST").json(), Message)

    def _reconnect(self):
        r = make_request(self.web + "gateway", method="GET")
        ws = websocket.create_connection(f"{r.json()['url']}/?v=9&encoding=json")
        js = {
            "op": 2,
            "d": {
                "token": self.token,
                "intents": 513,
                "properties": {
                    "$os": "windows",
                    "$browser": "opera",
                    "$device": "laptop"
                }
            }
        }
        ws.send(json.dumps(js))
        return ws

    def run(self, debug=False):
        print("* Starting bot and waiting for updates: DISCORD")
        if debug:
            print("* Debug mod: ENABLED")
        else:
            print("* Debug mod: DISABLED")

        ws = self._reconnect()
        while True:
            try:
                rc = json.loads(str(ws.recv()))
                if rc != "":
                    if rc["t"] == "MESSAGE_CREATE":
                        if rc["d"]["author"]["id"] != self.id:
                            if debug:
                                print(f"* New message: {convert_helper(rc['d'], Message)}")
                            for f in self.handle_message_funcs:
                                f(convert_helper(rc["d"], Message))
            except BaseException:
                ws = self._reconnect()
                continue
